# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU LIMS
#    Copyright (C) 2013  Luis Falcon <falcon@gnu.org>
#    Copyright (C) 2013  Sebastián Marró <smarro@thymbra.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from datetime import datetime
from trytond.model import ModelView, ModelSingleton, ModelSQL, fields
from trytond.pool import Pool


__all__ = ['GnuLimsSequences', 'SampleType', 'Sample', 'TestType', 'Lab',
    'GnuLimsTestUnits', 'GnuLimsTestCritearea', 'GnuLimsTest']


class GnuLimsSequences(ModelSingleton, ModelSQL, ModelView):
    "Standard Sequences for GnuLims"
    __name__ = "lims.sequences"

    lab_sequence = fields.Property(fields.Many2One('ir.sequence',
        'Lab Sequence', domain=[('code', '=', 'lims.lab')],
        required=True))
    lab_request_sequence = fields.Property(fields.Many2One('ir.sequence',
        'Lab Request Sequence', required=True,
        domain=[('code', '=', 'lims.test')]))


class SampleType(ModelSQL, ModelView):
    'Type of Sample'
    __name__ = 'lims.sample_type'

    name = fields.Char('Type', required=True)
    code = fields.Char('Code', required=True)
    info = fields.Text('Description')


class Sample(ModelSQL, ModelView):
    'Sample'
    __name__ = 'lims.sample'

    name = fields.Char('ID', help="Sample ID", readonly=True)
    barcode = fields.Char('Barcode')
    party = fields.Many2One('party.party', 'Party', required=True)
    external_sample = fields.Char('External Sample ID')
    type = fields.Many2One('lims.sample_type', 'Type')
    date_collected = fields.DateTime('Date Collected')
    date_expiration = fields.DateTime('Date Expiration')
    date_destroyed = fields.DateTime('Date Destroyed')
    date_frozen = fields.DateTime('Date Frozen')
    note = fields.Text('Note')


class TestType(ModelSQL, ModelView):
    'Type of Lab test'
    __name__ = 'lims.test_type'

    name = fields.Char('Test',
        help="Test type, eg X-Ray, hemogram,biopsy...", required=True,
        select=True)
    code = fields.Char('Code',
        help="Short name - code for the test", required=True, select=True)
    info = fields.Text('Description')
    product = fields.Many2One('product.product', 'Service', required=True)
    critearea = fields.One2Many('lims.test.critearea', 'test_type',
        'Test Cases')

    @classmethod
    def __setup__(cls):
        super(TestType, cls).__setup__()
        cls._sql_constraints = [
            ('code_uniq', 'unique(name)', 'The Lab Test code must be unique'),
        ]

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class Lab(ModelSQL, ModelView):
    'Lab Test'
    __name__ = 'lims.lab'

    name = fields.Char('ID', help="Lab result ID", readonly=True)
    test = fields.Many2One('lims.test_type', 'Test type',
        help="Lab test type", required=True, select=True)
    sample = fields.Many2One('lims.sample', 'Sample',
        required=True, select=True)
    results = fields.Text('Results')
    diagnosis = fields.Text('Diagnosis')
    critearea = fields.One2Many('lims.test.critearea',
        'lims_lab', 'Lab Test Critearea')
    date_requested = fields.DateTime('Date requested', required=True,
        select=True)
    date_analysis = fields.DateTime('Date of the Analysis', select=True)

    @classmethod
    def __setup__(cls):
        super(Lab, cls).__setup__()
        cls._sql_constraints += [
            ('id_uniq', 'unique (name)', 'The test ID code must be unique'),
        ]

    @staticmethod
    def default_date_requested():
        return datetime.now()

    @staticmethod
    def default_analysis():
        return datetime.now()

    @classmethod
    def create(cls, vlist):
        Sequence = Pool().get('ir.sequence')
        Config = Pool().get('lims.sequences')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('name'):
                config = Config(1)
                values['name'] = Sequence.get_id(
                    config.lab_sequence.id)

        return super(Lab, cls).create(vlist)


class GnuLimsTestUnits(ModelSQL, ModelView):
    'Test Units'
    __name__ = 'lims.test.units'

    name = fields.Char('Unit', select=True)
    code = fields.Char('Code', select=True)

    @classmethod
    def __setup__(cls):
        super(GnuLimsTestUnits, cls).__setup__()
        cls._sql_constraints = [
            ('name_uniq', 'unique(name)', 'The Unit name must be unique'),
        ]

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class GnuLimsTestCritearea(ModelSQL, ModelView):
    'Test Critearea'
    __name__ = 'lims.test.critearea'

    name = fields.Char('Analyte', required=True, select=True)
    excluded = fields.Boolean('Excluded', help='Select this option when'
        ' this analyte is excluded from the test')
    result = fields.Float('Value')
    result_text = fields.Char('Result - Text', help='Non-numeric results. For '
        'example qualitative values, morphological, colors ...')
    remarks = fields.Char('Remarks')
    normal_range = fields.Text('Reference')
    lower_limit = fields.Float('Lower Limit')
    upper_limit = fields.Float('Upper Limit')
    warning = fields.Boolean('Warn',
         on_change_with=['result', 'lower_limit', 'upper_limit'])
    units = fields.Many2One('lims.test.units', 'Units')
    test_type = fields.Many2One('lims.test_type', 'Test type',
     select=True)
    lims_lab = fields.Many2One('lims.lab', 'Test Cases',
        select=True)
    sequence = fields.Integer('Sequence')

    @classmethod
    def __setup__(cls):
        super(GnuLimsTestCritearea, cls).__setup__()
        cls._order.insert(0, ('sequence', 'ASC'))

    @staticmethod
    def default_sequence():
        return 1

    @staticmethod
    def default_excluded():
        return False

    def on_change_with_warning(self):
        if (self.result < self.lower_limit or self.result > self.upper_limit):
            return True
        return False

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class GnuLimsTest(ModelSQL, ModelView):
    'Lab Test'
    __name__ = 'lims.test'

    name = fields.Many2One('lims.test_type', 'Test Type',
        required=True, select=True)
    sample = fields.Many2One('lims.sample', 'Sample',
        required=True, select=True)
    date = fields.DateTime('Date', select=True)
    state = fields.Selection([
        ('draft', 'Draft'),
        ('tested', 'Tested'),
        ('ordered', 'Ordered'),
        ('cancel', 'Cancel'),
        ], 'State', readonly=True, select=True)
    request = fields.Integer('Request', readonly=True)
    urgent = fields.Boolean('Urgent')

    @classmethod
    def __setup__(cls):
        super(GnuLimsTest, cls).__setup__()
        cls._order.insert(0, ('date', 'DESC'))
        cls._order.insert(1, ('request', 'DESC'))
        cls._order.insert(2, ('name', 'ASC'))

    @staticmethod
    def default_date():
        return datetime.now()

    @staticmethod
    def default_state():
        return 'draft'

    @classmethod
    def create(cls, vlist):
        Sequence = Pool().get('ir.sequence')
        Config = Pool().get('lims.sequences')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('request'):
                config = Config(1)
                values['request'] = Sequence.get_id(
                    config.lab_request_sequence.id)

        return super(GnuLimsTest, cls).create(vlist)

    @classmethod
    def copy(cls, tests, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['request'] = None
        default['date'] = cls.default_date()
        return super(GnuLimsTest, cls).copy(tests,
            default=default)
